/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Mikhail Chupilko (chupilko@ispras.ru)
// SYNOPSIS: Reaction arbiters
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/arbiter/base.hpp>

namespace cpptesk {
namespace hw {

////////////////////////////////////////////////////////////////////////////////
// Two-Level Reaction Arbiter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_TWO_LEVEL_ARBITER ::cpptesk::hw::TwoLevelArbiter

#ifdef HW_SHORT_NAMES
    #define TWO_LEVEL_ARBITER CPPTESK_TWO_LEVEL_ARBITER
#endif // HW_SHORT_NAMES

template <class PrimaryArbiter, class SecondaryArbiter>
  class TwoLevelArbiter : public Arbiter
{
public:
    TwoLevelArbiter()
    {
        CPPTESK_ASSERTION(arb1.isModelBased() && !arb2.isModelBased(), "[SYSTEM] TwoLevelArbiter: A1 should be model based and A2 should be adaptive!");
    }
    virtual ~TwoLevelArbiter() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        arb1.appendRequest(process, message);
    }

    /// Add request to the list of requests passed first stage
    virtual bool preChooseRequest(const Process &process, const Message &message)
    {
        if (arb1.chooseRequest(process, message))
        {
            Request req = { (Process*)&process, (Message*)message.clone() };
            requests_from_first_stage.push_back(req);
        }
        else
            return false;

        return true;
    }

    virtual bool chooseRequest(const Process &process, const Message &message)
    {
        /*bool found = false;
        /// Look for the request in pre-selected; the the request is not found, return false
        for(unsigned int i = 0; i < requests_from_first_stage.size(); i++)
        {
            Request req = requests_from_first_stage[i];
            if(req.process == &process)
            {
                found = true;
                break;
            }
        }

        if(!found) return false;*/

        CPPTESK_ASSERTION(requests_from_first_stage.size(), "No requests were chosen by the first arbiter!");

        for(unsigned int i = 0; i < requests_from_first_stage.size(); i++)
        {
            Request req = requests_from_first_stage[i];
	    arb2.appendRequest((*req.process), (*req.message));
        }

        /// Ask arb2 to find our request in the list of pre-selected requests
	bool result = arb2.chooseRequest(process, message);

        /// Remove all requests from arb2
	for(unsigned int i = 0; i < requests_from_first_stage.size(); i++)
	{
	    Request req = requests_from_first_stage[i];
	    arb2.removeRequest(*(req.process), *(req.message));
	}

        /// Remove our request from the list of pre-selected requests
        for(unsigned int i = 0; i < requests_from_first_stage.size(); i++)
        {
            Request req = requests_from_first_stage[i];
            if(req.process == &process)
            {
                requests_from_first_stage.erase(requests_from_first_stage.begin() + i);
                break;
            }
        }

	return result;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        arb1.removeRequest(process, message);
    }
    
    virtual bool isModelBased() const { return arb1.isModelBased() && arb2.isModelBased(); }

    virtual std::vector<SharedMessage> getMessages() const { return arb1.getMessages(); }
    virtual ArbiterMsgIterator begin() { return arb1.begin(); }
    virtual ArbiterMsgIterator end() { return arb1.end(); }

protected:
    PrimaryArbiter arb1;
    SecondaryArbiter arb2;
    typedef struct Request { Process *process; SharedMessage message; } Request;
    /// Queue of reaction requests.
    std::vector<Request> requests_from_first_stage;

};

}} // namespace cpptesk::hw

