/** @file autofsm.hpp
 * This file contains automatical FSM generation class description and implementation.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <vector>

#include "ts/scenario.hpp"
#include "hw/model.hpp"

namespace cpptesk {
namespace hw {

/**
 * This class provides functionality of automatical generation of FSM from models.
 */
template<class UserModel>
class AutoFSM
{
public:
  /**
   * Constructor with reference to DUT model */
  AutoFSM(UserModel &dut) : dut(dut) { show_start = true; show_cycle = true; show_guard = true; };
  /**
   * Default virtual destructor */
  virtual ~AutoFSM() {};
  /**
   * Set parameter show_start */
  void set_show_start(bool show_start) { this->show_start = show_start; };
  /**
   * Set parameter show_cycle */
  void set_show_cycle(bool show_cycle) { this->show_cycle = show_cycle; };
  /**
   * Set parameter show_guard */
  void set_show_guard(bool show_guard) { this->show_guard = show_guard; };
  
  /**
   * Initialize AutoFSM */
  void initialize() { initial_state = get_state(); };
  
  /**
   * Return current state */
  std::string get_state()
  {
    const std::vector<Process::PointDescriptor> points = dut.getPipelineSlice();
    
    std::string state;
    std::stringstream out;
    
    int start_counter = 0, cycle_counter = 0, guard_counter = 0;
    
    for (unsigned int i = 0; i < points.size(); i++)
    {
        Process::PointDescriptor point = points[i];
    
        switch (point.type)
        {
            case START: if(show_start) { out << "{" << point.name << "}"; } else { start_counter++; } break;
            case CYCLE: if(show_cycle) { out << "{" << point.name << "}"; } else { cycle_counter++; } break;
            case GUARD: if(show_guard) { out << "{" << point.name << "}"; } else { guard_counter++; } break;
            default: CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_INFO, "Unknown type of point " <<  point.type << " in " << point.file << ":" << point.line);
        }
    }

    if(!show_start) out << "{" << start_counter << "}";
    if(!show_cycle) out << "{" << cycle_counter << "}";
    if(!show_guard) out << "{" << guard_counter << "}";

    state = out.str();

    CPPTESK_DEBUG_PRINTF(::cpptesk::hw::DEBUG_USER, "get_state(): %s", state.c_str());
    return state;
  };

  void scen_nop()
  {
      std::string state = get_state();
      dut.cycle();
      if(state != initial_state) while(state == get_state()) dut.cycle();
  };

protected:
  /**
   * Show labels START or only their number */
  bool show_start;
  /**
   * Show labels CYCLE or only their number */
  bool show_cycle;
  /**
   * Show labels WAIT or only their number */
  bool show_guard;
  /**
   * Reference to the DUT model */
  UserModel &dut;
  /**
   * State which is right after DUT reboot */
  std::string initial_state;
};

}}
