/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware interface
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/arbiter/arbiter.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>

#include <algorithm>
#include <ostream>
#include <string>
#include <vector>

namespace cpptesk {
namespace hw {

class Interface: public Component
{
public:
    enum interface_type_t
    {
        /// Input interface.
        INPUT  = 0x01,
        /// Output interface.
        OUTPUT = 0x02,
        /// Input/output interface (in particular, channel).
        INOUT  = INPUT | OUTPUT
    };

    Interface(interface_type_t type = INOUT, const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME):
        Component(name), _message(0), _type(type), _arbiter(0), _default_arbiter(Arbiter::createDefaultArbiter()), _request_count(0), _taken(false), _target(0) {}

    virtual ~Interface() { delete _default_arbiter; }

    void registerArbiter(Arbiter &arbiter) { _arbiter = &arbiter; }

    bool isInput()  const { return (_type & INPUT)  != 0; }
    bool isOutput() const { return (_type & OUTPUT) != 0; }

    // Basically, for input interfaces
    virtual void request(const Process &process, const Message &message);
    virtual void capture();
    virtual void release();

    const Message* getMessage() const { return _message; }

    // Basically, for output interfaces
    void appendRequest(const Process &process, const Message &message);
    bool chooseRequest(const Process &process, const Message &message);
    bool preChooseRequest(const Process &process, const Message &message);
    void removeRequest(const Process &process, const Message &message);

    Arbiter* getArbiter() { return _arbiter ? _arbiter : _default_arbiter; }

    int requestNumber() const { return _request_count; }

    /// Returns false if a reaction has been taken by someone else.
    bool takeReaction();
    bool isReactionTaken() const { return _taken; }

    void setTarget(const Interface &target) { _target = &target; }
    const Interface* getTarget() const { return _target; }

    /// Returns false if the connected arbiter is adaptive
    bool isModelBased() const;

protected:
    /// Pointer to a message being transmitted.
    const Message *_message;
    /// Interface type (input, output or input/output).
    interface_type_t _type;
    /// Reaction arbiter associated with the interface (for output interfaces only).
    Arbiter *_arbiter;
    /// Default adaptive arbiter.
    Arbiter *_default_arbiter;
    /// Number of requests competing for a reaction (for output interfaces only).
    int _request_count;
    /// Flag that shows if a reaction is taken or not (for output interfaces only).
    bool _taken;
    /// Target interface (for channels, especially broadcast ones).
    const Interface *_target;
};

class InputInterface: public Interface
{
public:
    InputInterface(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME): Interface(INPUT, name) {}
    virtual ~InputInterface() {}
};

class OutputInterface: public Interface
{
public:
    OutputInterface(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME): Interface(OUTPUT, name) {}
    virtual ~OutputInterface() {}
};

class InoutInterface: public Interface
{
public:
    InoutInterface(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME): Interface(INOUT, name) {}
    virtual ~InoutInterface() {}
};

inline std::ostream& operator << (std::ostream &out, const Interface &iface)
{
    return out << iface.getName();
}

}} // namespace cpptesk::hw
